/*
 * Copyright (C) 2002 Paul Martin <pm@debian.org>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/unistd.h>
#include <sys/mount.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/reboot.h>
#include <fcntl.h>
#include <ctype.h>


/* "Lazy" umount. requires kernel 2.4.11 or later */
#define MNT_DETACH      0x00000002

/* Use table of filesystems, as we haven't mounted /proc */
char *filesystems[] = {
    "reiserfs",
    "ext3",
    "ext2",
    NULL
};

/* Manpages lie... this isn't in linux/unistd.h */
extern int pivot_root(const char *new_root, const char *put_old);

void fdput(int fd, char *string)
{
    int length = strlen(string);

    write(fd, string, length);
}

#define putsn(a) fdput(1,(a))
#define puterr(a) {fdput(1,(a)); puts(strerror(errno));}


int main(int argc, char **argv, char **envp)
{
    int i;
    char buffer[80], *rootfs, *type, *p;
    ssize_t size;

    i = open("/rootfs", O_RDONLY);
    if (i < 0) {
	puterr("Can't open /rootfs: ");
	goto error_exit;
    }

    size = read(i, buffer, sizeof(buffer) - 1);
    close(i);

    if (size <= 0) {
	puterr("Error reading /rootfs: ");
	goto error_exit;
    }

    p = buffer;
    rootfs = type = NULL;
    while (isspace(*p))
	p++;
    rootfs = p;
    while ((*p) && !isspace(*p))
	p++;
    if (*p == '\0') {
	type = p;		/* default to auto */
    } else {
	*p++ = '\0';
	while (isspace(*p))
	    p++;
	type = p;
	while ((*p) && !isspace(*p))
	    p++;
	*p = '\0';
    }

    putsn("Mounting real root filesystem ");
    putsn(rootfs);

    if ((*type == '\0') || (strcmp(type, "auto") == 0)) {
	char **fs = filesystems;

	putsn(" (auto:");
	while (*fs != NULL) {
	    if (mount(rootfs, "/root", *fs, MS_RDONLY, NULL) == 0)
		break;
	    fs++;
	}
	if (*fs == NULL) {
	    puts("error)\nNo matching filesystems");
	    goto error_exit;
	}
	putsn(*fs);
	puts(")");
    } else {
	putsn(" (");
	putsn(type);
	puts(")");

	if (mount(rootfs, "/root", type, MS_RDONLY, NULL) != 0) {
	    puterr("mounting new rootfs: ");
	    goto error_exit;
	}
    }

    if (chdir("/root") != 0) {
	puterr("Error in chdir to /root: ");
	goto error_exit;
    }

    if (pivot_root(".", "boot") != 0) {
	puterr("Error in pivot_root: ");
	goto error_exit;
    }

    if (chroot(".") != 0) {
	puterr("Error in chroot: ");
	goto error_exit;
    }

    if (chdir("/") != 0) {
	puterr("Error in chdir to / : ");
	goto error_exit;
    }

    /* Note dev/console, *not* /dev/console */
    i = open("dev/console", O_RDWR, 0);
    if (i == -1) {
	puterr("Can't open /dev/console: ");
	goto error_exit;
    }
    dup2(i, 0);
    close(i);
    dup2(0, 1);
    dup2(0, 2);

    /* This lazy umount will only work with newer kernels */
    if (umount2("boot", MNT_DETACH) != 0) {
	puterr("Error in umounting /boot: ");
	/* goto error_exit; */
    }

    execve("/sbin/init", argv, envp);
    puterr("Error execing new /sbin/init: ");

  error_exit:
    sleep(15);
    reboot(RB_AUTOBOOT);

    return (0);
}
